package com.molino.ai;

/**
 * Implementacion generica del algoritmo Alpha-beta minimax .
 */
public abstract class MinMaxSearch<S extends Comparable>
{
    /**
	 * @uml.property  name="transpositionTable"
	 * @uml.associationEnd  
	 */
    private TranspositionTable<S> transpositionTable;
    private boolean interruptRequested;

    public MinMaxSearch()
    {
        super();
    }

    public MinMaxSearch(TranspositionTable<S> transpositionTable)
    {
        this.transpositionTable = transpositionTable;
    }

    public SearchNode<S> choose(Iterable<SearchNode<S>> successors, int depth)
    {
        interruptRequested = false;
        transpositionTable.reset();
        return chooseMax(successors, depth, -1000000000, +1000000000);
    }

    
    public void interrupt()
    {
        interruptRequested = true;
    }

    protected SearchNode<S> chooseMax(Iterable<SearchNode<S>> successors, int depth, int mybest, int herbest)
    {
        SearchNode<S> best = new SearchNode<S>(mybest);
        if ( interruptRequested ) return best;

        Iterable<SearchNode<S>> opponentSuccessors;
        SearchNode<S> attempt;

        for (SearchNode<S> successor : successors)
        {
            attempt = transpositionTable.get(successor);

            if (attempt == null)
            {
                opponentSuccessors = generateSuccessors(successor);
                
                if ( depth > 1 )
                    attempt = chooseMin(opponentSuccessors, depth-1, mybest, herbest);
                else
                    attempt = chooseMax(opponentSuccessors, mybest, herbest);
                
                if (attempt.state != null) 
                    transpositionTable.put(attempt);
            }

            if ( attempt.score > best.score )
            {
                best = successor;
                best.score = attempt.score;
            }

            if ( best.score > herbest ) 
                break;                      
        }

        return best;
    }

    protected SearchNode<S> chooseMax(Iterable<SearchNode<S>> successors, int mybest, int herbest)
    {        
        SearchNode<S> best = new SearchNode<S>(mybest);
        if ( interruptRequested ) return best;
        
        Iterable<SearchNode<S>> opponentSuccessors;
        SearchNode<S> attempt;

        for (SearchNode<S> successor : successors)
        {
            attempt = transpositionTable.get(successor);

            if (attempt == null)
            {
                successor.score = evaluate(successor);
                attempt=successor;
                transpositionTable.put(attempt);
            }

            if ( attempt.score > best.score )
            {
                best = successor;
                best.score = attempt.score;
            }

            if ( best.score > herbest ) 
                break;          
        }

        return best;
    }

    protected SearchNode<S> chooseMin(Iterable<SearchNode<S>> successors, int depth, int mybest, int herbest)
    {        
        SearchNode<S> worst = new SearchNode<S>(herbest);
        if ( interruptRequested ) return worst;

        Iterable<SearchNode<S>> opponentSuccessors;
        SearchNode<S> attempt;
        int score;

        for (SearchNode<S> successor : successors)
        {
            attempt = transpositionTable.get(successor);

            if (attempt == null)
            {
                opponentSuccessors = generateSuccessors(successor);
                
                if ( depth > 1 )
                    attempt = chooseMax(opponentSuccessors, depth-1, mybest, herbest);
                else                        
                    attempt = chooseMin(opponentSuccessors, mybest, herbest);                    
                
                if ( attempt.state != null ) 
                    transpositionTable.put(attempt);
            }

            if ( attempt.score < worst.score )
            {
                worst = successor;
                worst.score = attempt.score;
            }

            if ( worst.score < mybest ) 
                break;
        }

        return worst;
    }

    protected SearchNode<S> chooseMin(Iterable<SearchNode<S>> successors, int mybest, int herbest)
    {        
        SearchNode<S> worst = new SearchNode<S>(herbest);
        if ( interruptRequested ) return worst;

        Iterable<SearchNode<S>> opponentSuccessors;
        SearchNode<S> attempt;
        int score;

        for (SearchNode<S> successor : successors)
        {
            attempt = transpositionTable.get(successor);

            if (attempt == null)
            {
                successor.score = -evaluate(successor);
                attempt = successor;
                transpositionTable.put(attempt);
            }

            if ( attempt.score < worst.score )
            {
                worst = successor;
                worst.score = attempt.score;
            }

            if ( worst.score < mybest ) 
                break;
        }

        return worst;
    }

    protected abstract Iterable<SearchNode<S>> generateSuccessors(SearchNode<S> node);
    protected abstract int evaluate(SearchNode<S> node);
}
