package com.molino.mills.agent;

import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import com.molino.ai.MinMaxSearch;
import com.molino.ai.SearchNode;
import com.molino.mills.model.GameState;
import com.molino.mills.model.MoveHashMap;
import com.molino.mills.model.Player;
import com.molino.mills.model.move.Move;

public class ComputerPlayer extends Player
{
    static final Logger log = Logger.getLogger(ComputerPlayer.class);
    static final Logger stats = Logger.getLogger("stats");
    static final int MIN_DEPTH_BOUND = 2;
    static final int MAX_DEPTH_BOUND = 10;
    
    /**
	 * @uml.property  name="searchEngine"
	 * @uml.associationEnd  
	 */
    private MinMaxSearch<GameState> searchEngine;    
    private Level originalLogLevel;
    /**
	 * @uml.property  name="lastChoice"
	 * @uml.associationEnd  
	 */
    private SearchResult lastChoice;
    /**
	 * @uml.property  name="legalMoves"
	 * @uml.associationEnd  
	 */
    private MoveHashMap legalMoves;
    private Map<Move, SearchResult> preparedMoves;
    private boolean interruptSearch;
    
    /**
	 * @author   Emilio
	 */
    private enum WorkRequest {/**
	 * @uml.property  name="nONE"
	 * @uml.associationEnd  
	 */
    NONE, /**
	 * @uml.property  name="pREPARE"
	 * @uml.associationEnd  
	 */
    PREPARE, /**
	 * @uml.property  name="sEARCH"
	 * @uml.associationEnd  
	 */
    SEARCH, /**
	 * @uml.property  name="iNTERRUPT"
	 * @uml.associationEnd  
	 */
    INTERRUPT};
    /**
	 * @author   Emilio
	 */
    private enum WorkerState {/**
	 * @uml.property  name="iDLE"
	 * @uml.associationEnd  
	 */
    IDLE, /**
	 * @uml.property  name="pREPARING"
	 * @uml.associationEnd  
	 */
    PREPARING, /**
	 * @uml.property  name="sEARCHING"
	 * @uml.associationEnd  
	 */
    SEARCHING}; 
    /**
	 * @uml.property  name="workerState"
	 * @uml.associationEnd  
	 */
    private WorkerState workerState = WorkerState.IDLE;
    /**
	 * @uml.property  name="lastRequest"
	 * @uml.associationEnd  
	 */
    private WorkRequest lastRequest = WorkRequest.NONE;

    public ComputerPlayer()
    {
        super();      
        searchEngine = new MillSearch();
    }
    
    public void run()
    {
        while ( !isInterruptRequested() )
        {
            switch ( lastRequest )
            {
            case PREPARE:
                doPrepareSearch();
                break;
            case SEARCH:
                doMoveSearch();
                break;
            default:
                pause(50);
            }            
        }
    }
    
    @Override
    public void prepareNextMove(MoveHashMap legalMoves)
    {
        if ( !isIdle() )
            interruptSearch();
        
        requestTask(WorkRequest.PREPARE);
        this.legalMoves = legalMoves;
    }

    @Override
    public Move getNextMove(MoveHashMap legalMoves)
    {
        long startTime = System.currentTimeMillis();
        try
        {
            if ( !isIdle() )
                interruptSearch();
            
            this.legalMoves = legalMoves;
            
            requestTask(WorkRequest.SEARCH); 
            
            while ( !isMoveReady() )
                getGame().wait(50);
            

            long endTime = System.currentTimeMillis();
            long timeLeft = (getGame().getTimeLimit()*1000) - (endTime-startTime) - 75;
            
            if ( timeLeft > 0 )
                getGame().wait(timeLeft);
            
            interruptSearch();

            endTime = System.currentTimeMillis();
            log.info("Depth: "+lastChoice.depth+"\tTotal time: "+(endTime-startTime)+" ms");
        }
        catch (InterruptedException e)
        {
        }
        
        return popMove();
    }

    private synchronized void requestTask(WorkRequest request)
    {
        this.lastRequest = request;        
    }

    /**
     * 
     */
    private void interruptSearch()
    {
        searchEngine.interrupt();
        interruptSearch = true;
        while ( !isIdle() )
        try
        {
            getGame().wait(50);
        }
        catch (InterruptedException e)
        {
        }        
        interruptSearch = false;
    }

    private boolean isPreparing()
    {
        return workerState==WorkerState.PREPARING;
    }
    
    private boolean isSearching()
    {
        return workerState==WorkerState.SEARCHING;
    }
    
    private boolean isIdle()
    {
        return workerState==WorkerState.IDLE;
    }

    private boolean isMoveReady()
    {
        return lastChoice!=null && lastChoice.move!=null;
    }

    private Move popMove()
    {
        Move move = lastChoice.move;
        lastChoice = null;
        return move;
    }

    private synchronized void pause(int ms)
    {
        try
        {
            wait(ms);
        }
        catch (InterruptedException e)
        {
        }        
    }
    
    private synchronized void doPrepareSearch()
    {
        log.debug("Preparing for next turn...");
        workerState = WorkerState.PREPARING;
        BoardAnalyser.deactivateLog();


        preparedMoves = new HashMap<Move, SearchResult>(32);

        for (int db = 1; db <= MAX_DEPTH_BOUND; db++)
        {
            for (Move move : legalMoves.keySet())
            {
                if ( interruptSearch )
                    break;
                
                GameState state = legalMoves.get(move);
                

                MillSearchNodeMap moves = (MillSearchNodeMap) 
                    getGame().getMoveGenerator().generateMoves(new MillSearchNodeMap(32), state, false);
                
                SearchNode<GameState> choice =  searchEngine.choose(moves, db);
                
                if ( choice.state != null & !interruptSearch )
                {
                    preparedMoves.put(move, new SearchResult(choice.state.lastMove, db));
                }
                
                lastChoice = null;
            }
        }
        log.debug("Prepare moves done or aborted.");                    
        lastChoice = null;
        legalMoves = null;
        lastRequest = WorkRequest.NONE;
        workerState = WorkerState.IDLE;
    }

    /**
     * 
     */
    private synchronized void doMoveSearch()
    {
        log.debug("Searching...");
        workerState = WorkerState.SEARCHING;
        BoardAnalyser.deactivateLog();

        lastChoice = new SearchResult();
        int startingDepth = MIN_DEPTH_BOUND;        

        if ( preparedMoves != null  )
        {
            SearchResult c = preparedMoves.get(getGame().state.lastMove);
            if ( c!= null )
            {
                lastChoice = c;
                startingDepth = lastChoice.depth+1;
                log.debug("Continuing search from level "+startingDepth);
            }
        }
        
        MillSearchNodeMap moveMap = (new MillSearchNodeMap(legalMoves));
        
        for (int db = startingDepth; db <= MAX_DEPTH_BOUND; db++)
        {
            if ( interruptSearch )
                break;
            
            SearchNode<GameState> choice =  searchEngine.choose(moveMap, db);
            
            if ( choice.state != null & !interruptSearch )
            {
                lastChoice.depth = db;
                lastChoice.move = choice.state.lastMove;
            }            
        }
        
        log.debug("Search done or aborted.");                    
        BoardAnalyser.reactivateLog();                
        interruptSearch = false;
        legalMoves = null;
        preparedMoves = null;
        lastRequest = WorkRequest.NONE;
        workerState = WorkerState.IDLE;
   }
}

/**
 * @author  Emilio
 */
class SearchResult 
{
    public int depth;
    /**
	 * @uml.property  name="move"
	 * @uml.associationEnd  
	 */
    public Move move;
    
    public SearchResult()
    {
    }

    public SearchResult(Move move, int depth)
    {
        this.move = move;
        this.depth = depth;
    }

}