
package com.molino.mills.model;

import java.util.List;
import java.util.Vector;

import org.apache.log4j.Logger;

import com.molino.mills.agent.BoardAnalyser;
import com.molino.mills.agent.ComputerPlayer;
import com.molino.mills.model.event.GameEvent;
import com.molino.mills.model.event.GameOverEvent;
import com.molino.mills.model.event.InternalErrorEvent;
import com.molino.mills.model.event.MoveEvent;
import com.molino.mills.model.event.MovePromptEvent;
import com.molino.mills.model.event.TakeBackEvent;
import com.molino.mills.model.move.IllegalMoveException;
import com.molino.mills.model.move.Move;
import com.molino.mills.model.move.PassMove;


public class Game
    extends Thread
{
    static final Logger log = Logger.getLogger(Game.class); 
        
    /**
	 * @uml.property  name="state"
	 * @uml.associationEnd  
	 */
    public GameState state;
    /**
	 * @uml.property  name="players"
	 * @uml.associationEnd  multiplicity="(0 -1)"
	 */
    private Player[] players = new Player[2];
    private List<GameListener> gameListeners;
    /**
	 * @uml.property  name="moveGenerator"
	 * @uml.associationEnd  
	 */
    private MoveGenerator moveGenerator = new MoveGenerator();
    /**
	 * @uml.property  name="legalMoves"
	 * @uml.associationEnd  
	 */
    private MoveHashMap legalMoves;
    /**
	 * @uml.property  name="winner"
	 * @uml.associationEnd  
	 */
    private Player winner = null;
    private boolean interruptRequested = false;
    private boolean takeBackRequested;
    /**
	 * @uml.property  name="timeLimit"
	 */
    private int timeLimit;    
    
    public Game(Player[] players)
    {
        this(players[0], players[1]);
    }
    
    public Game(Player player1, Player player2)
    {
        gameListeners = new Vector<GameListener>();
        state = new GameState();
        players[0] = player1;
        players[0].setGame(this);
        players[0].setPlayerNumber(0);
        players[1] = player2;
        players[1].setGame(this);
        players[1].setPlayerNumber(1);
    }
    
    public void run()
    {
        play();        
    }
    
    public synchronized void play()
    {
        int turn;
        Player player, opponent;
        Move move = null;
        GameState newState;
        try
        {
            players[0].start();
            players[1].start();
            
            while (!interruptRequested)       
            {
                turn = state.currentPlayer;
                player = players[turn];
                opponent = players[1-turn];
                log.debug("--------------------------------------------------------------------------------------------------------------------------------------------------");
                log.info(player+"'s turn (turnsCompleted="+state.turnsCompleted+")");
                log.info("Pieces left: "+state.board.piecesLeft[turn]+"\tPieces to place: "+state.board.unplacedPieces[turn]);
                log.info("Stage: "+state.stage+"\tCapture(s) pending: "+state.awaitingCapture);
                
                if ( log.isDebugEnabled() )
                    BoardAnalyser.logBoardEvaluation(state);
                
                legalMoves = (MoveHashMap) moveGenerator.generateMoves(new MoveHashMap(23), state, true);
                log.debug("Next possibile moves: "+legalMoves.keySet());
                
                
                if ( isGameOver() )
                {
                    winner = players[1-turn];
                    log.info("Game over!  Winner: "+winner);
                    notifyListeners(new GameOverEvent(this));
                    return;
                }

                MovePromptEvent movePromptEvent = new MovePromptEvent(this, state, player, state.stage, state.awaitingCapture);
                notifyListeners(movePromptEvent);
                
 
                opponent.prepareNextMove(legalMoves);
                
                
                move = player.getNextMove(legalMoves); 
                
                if ( takeBackRequested )
                {
                    takeBackRequested = false;
                    getCurrentPlayer().clearInterruptRequest();
                    takeBackMove();
                }
                else if ( interruptRequested )
                {
                    break;
                }
                else
                {
                    processMoveRequest(player, move);
                }
            }
        } 
        catch (Throwable e)
        {
            notifyListeners(new InternalErrorEvent(this, e));                
        }
    }

    /**
     * @param player
     * @param move
     */
    private void processMoveRequest(Player player, Move move)
    {
        GameState newState;
        log.info(player+" requests move: "+move);
        
        newState = legalMoves.get(move);
        
        if ( newState==null )
        {
            notifyListeners(new InternalErrorEvent(this, new IllegalMoveException(move)));
        }
        else
        {
            MoveEvent moveEvent = new MoveEvent(this, state.turnsCompleted+1, player, move);
        
            this.state = newState;
            

            
            notifyListeners(moveEvent);
        }
    }

    /**
     * 
     */
    private void takeBackMove()
    {
        if ( state.previous == null )
        {
            notifyListeners(new TakeBackEvent(this, false));            
            log.info("Move undo request rejected.");
        }
        else
        {
            state = state.previous; 
            if ( state.awaitingCapture > 0 )
                state = state.previous;
            notifyListeners(new TakeBackEvent(this, true));
            if ( state.previous!=null && getPlayer(state.currentPlayer) instanceof ComputerPlayer )
            {
                state = state.previous;
                notifyListeners(new TakeBackEvent(this, true));
            }
            log.info("Move taken back.");
        }
    }
    
    public boolean isPassRequired()
    {
        return legalMoves.size()==1 && legalMoves.keySet().iterator().next() instanceof PassMove;
    }

    public boolean isGameOver()
    {        
        return legalMoves.size()==0 || state.board.piecesLeft[state.currentPlayer]<3;
    }


    public Player getPlayer(int i)
    {
        return players[i];
    }
    
    public void addGameListener(GameListener gl)
    {
        gameListeners.add(gl);
    }
    
    public void notifyListeners(GameEvent event)
    {
        for (GameListener gl : gameListeners)
        {
            gl.onGameEvent(event);
        }
    }

    public boolean isLegalMove(Move move)
    {       
        return legalMoves.containsKey(move);
    }

    /**
	 * @return
	 * @uml.property  name="moveGenerator"
	 */
    public MoveGenerator getMoveGenerator()
    {
        return moveGenerator;
    }

    /**
	 * @return
	 * @uml.property  name="winner"
	 */
    public Player getWinner()
    {
        return winner;
    }

    /**
	 * @return
	 * @uml.property  name="legalMoves"
	 */
    public MoveHashMap getLegalMoves()
    {
        return legalMoves;
    }

    public void interrupt()
    {
        interruptRequested = true;
        players[0].interrupt();
        players[1].interrupt();
        
        while ( isAlive() )
        {
        }
    }

    public void requestTakeBackMove()
    {
        takeBackRequested = true;
        getCurrentPlayer().interrupt();
    }

    public Player getCurrentPlayer()
    {
        return getPlayer(state.currentPlayer);
    }

    /**
	 * @param i
	 * @uml.property  name="timeLimit"
	 */
    public void setTimeLimit(int i)
    {
        this.timeLimit = i;       
    }

    /**
	 * @return
	 * @uml.property  name="timeLimit"
	 */
    public int getTimeLimit()
    {
        return timeLimit;
    }    
}
