package com.molino.mills.model;

import com.molino.mills.model.move.CaptureMove;
import com.molino.mills.model.move.FromToMove;
import com.molino.mills.model.move.Move;
import com.molino.mills.model.move.MoveAndCaptureMove;
import com.molino.mills.model.move.PlaceMove;

public class MoveGenerator
{
    public static final int[][] ADJACENT_POSITIONS =
    {

        {7, 1},             // 0
        {0, 2, 9},          // 1
        {1, 3},             // 2
        {2, 4, 11},         // 3
        {3, 5},             // 4
        {4, 6, 13},         // 5
        {5, 7},             // 6
        {6, 0, 15},         // 7
        

        {15,  9},           // 8
        { 8, 10, 1, 17},    // 9
        { 9, 11},           // 10
        {10, 12, 3, 19},    // 11
        {11, 13},           // 12
        {12, 14, 5, 21},    // 13
        {13, 15},           // 14
        {14,  8, 7, 23},    // 15
    

        {23, 17},           // 16
        {16, 18, 9},        // 17
        {17, 19},           // 18
        {18, 20, 11},       // 19
        {19, 21},           // 20
        {20, 22, 13},       // 21
        {21, 23},           // 22
        {22, 16, 15}        // 23
    };
    
    public static final int[] ADJACENT_POSITION_BBS = BitBoardUtil.positionsArrayToBitBoards(ADJACENT_POSITIONS);

    public MoveGenerator()
    {
    }
    
    public MoveMap generateMoves(MoveMap moves, GameState state, boolean includeIntermediate)
    {
        if ( state.awaitingCapture > 0)
            return generateCaptureMoves(moves, state);
        else
        switch (state.stage)
        {
        case OPENING:
            return generatePlaceMoves(moves, state, includeIntermediate);
        case MIDGAME:
            return generateSlideMoves(moves, state, includeIntermediate);
        case ENDGAME: 
        default:
            return generateJumpMoves(moves, state, includeIntermediate);
        }
    }

    protected MoveMap generateCaptureMoves(MoveMap moves, GameState state)
    {
    
        Board board = state.board;
        int player = state.currentPlayer;
        int opponent = 1-player;
        int playerbb = board.bbs[player];
        int opponentbb = board.bbs[opponent];
        int[] unplacedPieces = board.unplacedPieces;
        int[] piecesLeft = board.piecesLeft;
        
    
        int[] newbbs;
        
    
        int captureablebb = opponentbb & ~state.features.millbbs[opponent];
        
        if ( captureablebb == 0 )
            captureablebb = opponentbb;            
        
    
        int[] newPiecesLeft = new int[2];
        newPiecesLeft[player] = piecesLeft[player];
        newPiecesLeft[opponent] = piecesLeft[opponent] - 1;

    
        for (int bb = captureablebb, mask = 1, pos = 0; 
             bb != 0; 
             bb >>>= 1, mask <<= 1, pos++)
            if ((bb & 1) != 0)
            {
                newbbs = new int[2];
                newbbs[player] = playerbb;
                newbbs[opponent] = opponentbb ^ mask;
                moves.put(new GameState(new Board(newbbs, unplacedPieces, newPiecesLeft), state, new CaptureMove(pos)));
            }
        
        return moves;
    }
    
    protected MoveMap generatePlaceMoves(MoveMap moves, GameState state, boolean includeIntermediate)
    {
    
        int player = state.currentPlayer;
        Board board = state.board;
        int opponent = 1-player;
        int playerbb = board.bbs[player];
        int opponentbb = board.bbs[opponent];
        int[] unplacedPieces = board.unplacedPieces;
        int[] piecesLeft = board.piecesLeft;
        int freebb = state.features.freebb;  
        
    
        int[] newbbs;
        
     
        int[] newUnplacedPieces = new int[2];
        newUnplacedPieces[player] = unplacedPieces[player]-1;
        newUnplacedPieces[opponent] = unplacedPieces[opponent]; 

        for ( int bb=freebb, mask=1, pos=0; 
              bb!=0; 
              bb>>>=1, mask<<=1, pos++ )
        if ( (bb & 1)!=0 )
        {
            newbbs = new int[2];
            newbbs[player] = playerbb ^ mask;
            newbbs[opponent] = opponentbb;
            
            expandAndPut(moves, new GameState(new Board(newbbs, newUnplacedPieces, piecesLeft), state, new PlaceMove(pos)), 
                        includeIntermediate);
        }

        return moves;
    }

    private void expandAndPut(MoveMap moves, GameState state, boolean includeIntermediate)
    {
        if ( state.awaitingCapture==0 )
        {
            moves.put(state);
        }
        else
        {
            if ( includeIntermediate )
                moves.put(state);
            
            expandCaptureMoves(moves, state);
        }
    }

    private void expandCaptureMoves(MoveMap moves, GameState state)
    {
        MoveHashMap captureMoveMap = new MoveHashMap(9);
        generateCaptureMoves(captureMoveMap, state);
        GameState capturedState;
        MoveAndCaptureMove macMove;
        for (Move captureMove : captureMoveMap.keySet())
        {
            capturedState = captureMoveMap.get(captureMove);
            capturedState.lastMove = new MoveAndCaptureMove(state.lastMove, (CaptureMove) captureMove);
            moves.put(capturedState);
        }
    }    
    
    protected MoveMap generateSlideMoves(MoveMap moves, GameState state, boolean includeIntermediate)
    {
    
        int player = state.currentPlayer;
        Board board = state.board;
        int opponent = 1-player;
        int playerbb = board.bbs[player];
        int opponentbb = board.bbs[opponent];
        int[] unplacedPieces = board.unplacedPieces;
        int[] piecesLeft = board.piecesLeft;
        int freebb = state.features.freebb;  

    
        int[] newbbs;
        
  
        for ( int bb=playerbb, pos=0, mask=1; 
              bb!=0; 
              bb>>>=1, mask<<=1, pos++ )
        if ( (bb & 1)!=0 )
        {
  
            int adjbb = MoveGenerator.ADJACENT_POSITION_BBS[pos] & freebb;
            
 
            for (int destMask=1, dest=0; 
                 adjbb!=0; 
                 adjbb>>>=1, destMask<<=1, dest++)
            if ( (adjbb & 1) != 0 )
            {
                newbbs = new int[2];
                newbbs[player] = playerbb ^ (mask | destMask);
                newbbs[opponent] = opponentbb;
                expandAndPut(moves,  
                        new GameState(new Board(newbbs, unplacedPieces, piecesLeft), state, 
                                      new FromToMove(pos, dest)), 
                        includeIntermediate);
            }
        }            
        
        return moves;
    }

    protected MoveMap generateJumpMoves(MoveMap moves, GameState state, boolean includeIntermediate)
    {
 
        int player = state.currentPlayer;
        Board board = state.board;
        int opponent = 1-player;
        int playerbb = board.bbs[player];
        int opponentbb = board.bbs[opponent];
        int[] unplacedPieces = board.unplacedPieces;
        int[] piecesLeft = board.piecesLeft;
        int freebb = state.features.freebb;  
        
 
        Board newBoard;
        BoardFeatures newFeatures;
        GameState newState;
        int[] newbbs;
        Move move;


        for (int bb = playerbb, pos = 0, mask = 1; 
             bb != 0; 
             bb >>>= 1, mask <<= 1, pos++)
        if ((bb & 1) != 0)
        {

            for (int bb2 = freebb, destMask = 1, dest = 0; 
                 bb2 != 0; 
                 bb2 >>>= 1, destMask <<= 1, dest++)
            if ((bb2 & 1) != 0)
            {
                newbbs = new int[2];
                newbbs[player] = playerbb ^ (mask | destMask);
                newbbs[opponent] = opponentbb;
                
                expandAndPut(moves, 
                        new GameState(new Board(newbbs, unplacedPieces, piecesLeft), state, 
                                      new FromToMove(pos, dest)), 
                        includeIntermediate);
            }
        }
        
        return moves;
    }    
}
